@@ -11,6 +11,7 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+//go:build !nozfs
 // +build !nozfs
 
 package collector
@@ -19,7 +20,10 @@ import (
 	"fmt"
 
 	"github.com/go-kit/log"
+	"github.com/go-kit/log/level"
 	"github.com/prometheus/client_golang/prometheus"
+
+	"golang.org/x/sys/unix"
 )
 
 type zfsCollector struct {
@@ -44,6 +48,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.abdstats.linear_cnt",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 			{
 				name:        "abdstats_linear_data_bytes",
@@ -51,6 +56,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.abdstats.linear_data_size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "abdstats_scatter_chunk_waste_bytes",
@@ -58,6 +64,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.abdstats.scatter_chunk_waste",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "abdstats_scatter_count_total",
@@ -65,6 +72,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.abdstats.scatter_cnt",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 			{
 				name:        "abdstats_scatter_data_bytes",
@@ -72,6 +80,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.abdstats.scatter_data_size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "abdstats_struct_bytes",
@@ -79,6 +88,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.abdstats.struct_size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_anon_bytes",
@@ -86,6 +96,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.anon_size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_c_bytes",
@@ -93,6 +104,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.c",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_c_max_bytes",
@@ -100,6 +112,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.c_max",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_c_min_bytes",
@@ -107,6 +120,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.c_min",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_data_bytes",
@@ -114,6 +128,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.data_size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_demand_data_hits_total",
@@ -121,6 +136,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.demand_data_hits",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_demand_data_misses_total",
@@ -128,6 +144,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.demand_data_misses",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_demand_metadata_hits_total",
@@ -135,6 +152,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.demand_metadata_hits",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_demand_metadata_misses_total",
@@ -142,6 +160,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.demand_metadata_misses",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_hdr_bytes",
@@ -149,6 +168,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.hdr_size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_hits_total",
@@ -156,6 +176,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.hits",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_misses_total",
@@ -163,6 +184,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.misses",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_mfu_ghost_hits_total",
@@ -170,6 +192,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.mfu_ghost_hits",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_mfu_ghost_size",
@@ -177,6 +200,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.mfu_ghost_size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_mfu_bytes",
@@ -184,6 +208,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.mfu_size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_mru_ghost_hits_total",
@@ -191,6 +216,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.mru_ghost_hits",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_mru_ghost_bytes",
@@ -198,6 +224,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.mru_ghost_size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_mru_bytes",
@@ -205,6 +232,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.mru_size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "arcstats_other_bytes",
@@ -212,13 +240,37 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.other_size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
+			// when FreeBSD 14.0+, `meta/pm/pd` install of `p`.
 			{
 				name:        "arcstats_p_bytes",
 				description: "ZFS ARC MRU target size",
 				mib:         "kstat.zfs.misc.arcstats.p",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
+			},
+			{
+				name:        "arcstats_meta_bytes",
+				description: "ZFS ARC metadata target frac ",
+				mib:         "kstat.zfs.misc.arcstats.meta",
+				dataType:    bsdSysctlTypeUint64,
+				valueType:   prometheus.GaugeValue,
+			},
+			{
+				name:        "arcstats_pd_bytes",
+				description: "ZFS ARC data MRU target frac",
+				mib:         "kstat.zfs.misc.arcstats.pd",
+				dataType:    bsdSysctlTypeUint64,
+				valueType:   prometheus.GaugeValue,
+			},
+			{
+				name:        "arcstats_pm_bytes",
+				description: "ZFS ARC meta MRU target frac",
+				mib:         "kstat.zfs.misc.arcstats.pm",
+				dataType:    bsdSysctlTypeUint64,
+				valueType:   prometheus.GaugeValue,
 			},
 			{
 				name:        "arcstats_size_bytes",
@@ -226,6 +278,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.arcstats.size",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.GaugeValue,
+				labels:      nil,
 			},
 			{
 				name:        "zfetchstats_hits_total",
@@ -233,6 +286,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.zfetchstats.hits",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 			{
 				name:        "zfetchstats_misses_total",
@@ -240,6 +294,7 @@ func NewZfsCollector(logger log.Logger) (Collector, error) {
 				mib:         "kstat.zfs.misc.zfetchstats.misses",
 				dataType:    bsdSysctlTypeUint64,
 				valueType:   prometheus.CounterValue,
+				labels:      nil,
 			},
 		},
 		logger: logger,
@@ -250,7 +305,9 @@ func (c *zfsCollector) Update(ch chan<- prometheus.Metric) error {
 	for _, m := range c.sysctls {
 		v, err := m.Value()
 		if err != nil {
-			return fmt.Errorf("couldn't get sysctl: %w", err)
+			// debug logging
+			level.Debug(c.logger).Log("name", m.name, "couldn't get sysctl:", err)
+			continue
 		}
 
 		ch <- prometheus.MustNewConstMetric(
@@ -263,3 +320,42 @@ func (c *zfsCollector) Update(ch chan<- prometheus.Metric) error {
 
 	return nil
 }
+
+func (c *zfsCollector) parseFreeBSDPoolObjsetStats() error {
+
+	sysCtlMetrics := []string{
+		"nunlinked", "nunlinks", "nread", "reads", "nwritten", "writes",
+	}
+	zfsPoolMibPrefix := "kstat.zfs.pool.dataset"
+	zfsDatasetNames := []string{}
+
+	zfsDatasets, err := unix.Sysctl(zfsPoolMibPrefix)
+	if err != nil {
+		return fmt.Errorf("couldn't get sysctl: %w", err)
+	}
+
+	for dataset, _ := range zfsDatasets {
+		if strings.HasSuffix(dataset, ".dataset_name") {
+			zfsDatasetNames = append(zfsDatasetNames, strings.SplitAfter(dataset, ".")[3])
+		}
+	}
+
+	for zpoolDataset := range zfsDatasetsNames {
+		zfsDatasetLabels := map[string]string{
+			"dataset": zpoolDataset,
+			"zpool":   strings.SplitAfter(zpoolDataset, "/")[0],
+		}
+		for metric := range sysCtlMetrics {
+			c.sysctls = append(c.sysctls, bsdSysctl{
+				name:        fmt.SprintF("node_zfs_zpool_dataset_%s", metric),
+				description: fmt.SprintF("node_zfs_zpool_dataset_%s", metric),
+				mib:         fmt.Sprintf("%s.%s.%s", zfsPoolMibPrefix, poolObj, metric),
+				dataType:    bsdSysctlTypeUint64,
+				valueType:   prometheus.CounterValue,
+				labels:      zfsDatasetLabels,
+			})
+		}
+	}
+
+	return nil
+}
